/*
 * EFI Boot Guard
 *
 * Copyright (c) Siemens AG, 2017
 *
 * Authors:
 *  Andreas Reichel <andreas.reichel.ext@siemens.com>
 *
 * This work is licensed under the terms of the GNU GPL, version 2.  See
 * the COPYING file in the top-level directory.
 *
 * SPDX-License-Identifier:	GPL-2.0-only
 */

#include "bootguard.h"
#include "utils.h"

EFI_HANDLE this_image;

VOLUME_DESC *volumes = NULL;
UINTN volume_count = 128;
CHAR16 *boot_medium_path;
